#include <cstdio>
#include <cstring>
#include <iconv.h>
#include <unistd.h>
#include <pqxx/pqxx>
#include <SDL.h>

#include "flagtrigger.h"
#include "widestring.h"
#include "fetch_current_tournament.h"
#include "fetch_list_of_active_groups.h"
#include "fetch_list_of_finished_groups.h"
#include "fetch_group.h"
#include "fetch_auxilliary_screens.h"
#include "fonts.h"
#include "groupscreen.h"
#include "top10scorescreen.h"
#include "top5chosenscreen.h"
#include "splitscreen.h"
#include "rotatescreen.h"

SDL_Surface *screen = NULL;

Tournament active_tournament;
std::vector<SkeletonGroup> active_groups;
std::vector<GenericScreen *> screens;
GenericScreen *mainscreen = NULL;
unsigned char framebuf[SCREEN_WIDTH * SCREEN_HEIGHT * 4], screenbuf[SCREEN_WIDTH * SCREEN_HEIGHT * 4];

void init(pqxx::connection &conn)
{
	std::vector<widestring> aux_screens;
	
	if (screens.size() == 0 || mainscreen != screens[0])
		delete mainscreen;
	
	for (std::vector<GenericScreen *>::const_iterator i = screens.begin(); i != screens.end(); ++i) {
		delete *i;
	}
	screens.erase(screens.begin(), screens.end());

	bool show_only_main_screen = (USE_SPLITSCREEN && screens.size() == 1);

#if !USE_SPLITSCREEN
	RotateScreen *rs = new RotateScreen();
	mainscreen = rs;
#endif

	conn.perform(FetchCurrentTournament(&active_tournament));
	conn.perform(FetchListOfActiveGroups(&active_groups));

	if (active_tournament.id == -1) {
		std::fprintf(stderr, "No active tournament\n");
	} else {
		std::fprintf(stderr, "Current tournament is %d\n", active_tournament.id);

		for (std::vector<SkeletonGroup>::const_iterator i = active_groups.begin(); i != active_groups.end(); ++i) {
			std::fprintf(stderr, "tourn: %u  round: %u  parallel: %u  num_machines: %u\n",
				i->tournament, i->round, i->parallel, i->num_machines);

			// memory leaks here?
			for (unsigned j = 0; j < i->num_machines; ++j) {
#if USE_SPLITSCREEN
				RotateScreen *rs = new RotateScreen();
				screens.push_back(rs);
#endif
				rs->add_screen(new GroupScreen(conn, i->tournament, i->round, i->parallel, j, i->num_machines, i->players_per_machine));
			}
		}
	}

	/*
	 * Show auxilliary screens except if we have too many already,
	 * or if we're in the special split-screen end-tournament mode,
	 * where there if only one.
	 */
	RotateScreen *aux_screen = NULL;
	if (screens.size() < 4 && !show_only_main_screen) {
#if USE_SPLITSCREEN
		RotateScreen *rs = new RotateScreen();
		screens.push_back(rs);
#endif
		aux_screen = rs;

		conn.perform(FetchAuxilliaryScreens(&aux_screens));
		for (std::vector<widestring>::const_iterator i = aux_screens.begin(); i != aux_screens.end(); ++i) {
			if (*i == widestring("top10scores")) {
				rs->add_screen(new Top10ScoreScreen(conn, active_tournament.id));
				continue;
			}
			if (*i == widestring("top5chosen")) {
				rs->add_screen(new Top5ChosenScreen(conn, active_tournament.id));
				continue;
			}
		}
	}

#if USE_SPLITSCREEN
	/*
	 * If we still have room, make yet another rotational screen with
	 * results from previous groups -- otherwise tack them onto the end
	 * of the auxilliary screens.
	 */
	RotateScreen *finished_groups_screen;
	if (show_only_main_screen) {
		finished_groups_screen = NULL;
	} else if (screens.size() < 4) {
		finished_groups_screen = new RotateScreen();
		screens.push_back(finished_groups_screen);
	} else {
		finished_groups_screen = aux_screen;
	}
	if (finished_groups_screen != NULL) {
		std::vector<SkeletonGroup> finished_groups;
		conn.perform(FetchListOfFinishedGroups(active_tournament.id, &finished_groups));

		for (std::vector<SkeletonGroup>::const_iterator i = finished_groups.begin(); i != finished_groups.end(); ++i) {
			finished_groups_screen->add_screen(new GroupScreen(conn, i->tournament, i->round, i->parallel, 0, 1, 1));
		}
	}
#endif

#if USE_SPLITSCREEN
	// hack
	screens.push_back(NULL);
	screens.push_back(NULL);
	screens.push_back(NULL);
	screens.push_back(NULL);

	if (screens[1] == NULL) {
		mainscreen = screens[0];
	} else {
		mainscreen = new SplitScreen(screens[0], screens[1], screens[2], screens[3]);
	}
#endif
}

void main_loop(pqxx::connection &conn)
{
	if (active_tournament.id == -1) {
		// No active tournament, sleep a second or so and exit
		conn.await_notification(1, 0);
		return;
	}

	if (mainscreen && mainscreen->check_invalidated()) {
		if (screen->pitch == SCREEN_WIDTH) {
			SDL_LockSurface(screen);
			mainscreen->draw((unsigned char *)screen->pixels, SCREEN_WIDTH, SCREEN_HEIGHT);
			SDL_UnlockSurface(screen);
		} else {
			mainscreen->draw(framebuf, SCREEN_WIDTH, SCREEN_HEIGHT);
			SDL_LockSurface(screen);
			for (unsigned y = 0; y < SCREEN_HEIGHT; ++y) {
				unsigned char *sptr = framebuf + y * SCREEN_WIDTH * 4;
				unsigned char *dptr = (unsigned char *)screen->pixels + y * screen->pitch;
				memcpy(dptr, sptr, SCREEN_WIDTH * 4);
			}
			SDL_UnlockSurface(screen);
		}
		SDL_Flip(screen);
		conn.await_notification(0, 10000);
	} else {
		SDL_Flip(screen);
		conn.await_notification(0, 200000);
	}
}

int main(int argc, char **argv)
{
	SDL_Init(SDL_INIT_VIDEO);
#if USE_FULLSCREEN
	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, SDL_DOUBLEBUF | SDL_FULLSCREEN);
	SDL_ShowCursor(SDL_DISABLE);
#else
	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, SDL_DOUBLEBUF);
#endif
	if (screen == NULL) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(1);
	}
	
	try {
		init_freetype();
		pqxx::connection conn("dbname=ccbs host=www.positivegaming.com user=ccbs password=GeT|>>B_");
		FlagTrigger tournament_changed(conn, "active_tournament");
		FlagTrigger rounds_changed(conn, "active_groups");
		
		// when active_tournament or active_rounds is changed, we destroy everything and start from scratch
		// (at least currently)
		for ( ;; ) {
			tournament_changed.reset_flag();
			rounds_changed.reset_flag();
			init(conn);
			do {
				main_loop(conn);
				conn.get_notifs();
			} while (!tournament_changed.get_flag() && !rounds_changed.get_flag());
			std::fprintf(stderr, "active_tournament or active_groups changed, resetting...\n");
		}
	} catch (const std::exception &e) {
		std::fprintf(stderr, "Exception: %s\n", e.what());
		exit(1);
	}

	SDL_Quit();
	
	return 0;
}
